# (c) 2007 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

'''Abstract user interface, which provides all logic and strings.

Concrete implementations need to implement a set of abstract presentation
functions with an appropriate toolkit.
'''

import gettext, optparse, textwrap, urllib2, tempfile, sys, time, os

import gobject
import dbus
import dbus.service
import dbus.mainloop.glib

from oslib import OSLib
from backend import UnknownHandlerException, PermissionDeniedByPolicy, polkit_auth_wrapper, dbus_sync_call_signal_wrapper, Backend, DBUS_BUS_NAME

def bool(str):
    '''Convert backend encoding of a boolean to a real boolean.'''

    if str == 'True':
        return True
    assert str == 'False'
    return False

class AbstractUI(dbus.service.Object):
    '''Abstract user interface.

    This encapsulates the entire program logic and all strings, but does not
    implement any concrete user interface.
    '''
    def __init__(self):
        '''Initialize system.
        
        This parses command line arguments, detects available hardware,
        and already installed drivers and handlers.
        '''
        gettext.install('jockey', unicode=True)

        (self.argv_options, self.argv_args) = self.parse_argv()

        if self.argv_options.check:
            time.sleep(self.argv_options.check)

        self.init_strings()

        self._dbus_iface = None

        self.dbus_server_main_loop = None

    def backend(self):
        '''Return D-BUS backend client interface.

        This gets initialized lazily.
        '''
        if self._dbus_iface is None:
            self._dbus_iface = Backend.create_dbus_client()
        else:
            # handle backend timeouts
            try:
                self._dbus_iface.handler_info(' ')
            except Exception, e:
                if hasattr(e, '_dbus_error_name') and e._dbus_error_name == \
                    'org.freedesktop.DBus.Error.ServiceUnknown':
                    self._dbus_iface = Backend.create_dbus_client()

        return self._dbus_iface

    def _(self, str, convert_keybindings=False):
        '''Keyboard accelerator aware gettext() wrapper.
        
        This optionally converts keyboard accelerators to the appropriate
        format for the frontend.

        All strings in the source code should use the '_' prefix for key
        accelerators (like in GTK). For inserting a real '_', use '__'.
        '''
        result = _(str)

        if convert_keybindings:
            result = self.convert_keybindings(result)

        return result

    def init_strings(self):
        '''Initialize all static strings which are used in UI implementations.'''

        self.string_handler = self._('Component')
        self.string_enabled = self._('Enabled')
        self.string_disabled = self._('Disabled')
        self.string_status = self._('Status')
        self.string_restart = self._('Needs computer restart')
        self.string_in_use = self._('In use')
        self.string_not_in_use = self._('Not in use')
        # this is used in --list to denote free/restricted drivers
        self.string_free = self._('free')
        # this is used in --list to denote free/restricted drivers
        self.string_restricted = self._('restricted')
        self.string_download_progress_title = self._('Download in progress')
        self.string_unknown_driver = self._('Unknown driver')
        self.string_unprivileged = self._('You are not authorized to perform this action.')

    def main_window_title(self):
        '''Return an appropriate translated window title.

        This might depend on the mode the program is called (e. g. showing only
        free drivers, only restricted ones, etc.).
        '''
        if self.argv_options.mode == 'nonfree':
            return self._('Restricted Hardware Drivers')
        else:
            return self._('Hardware Drivers')

    def main_window_text(self):
        '''Return a tuple (heading, subtext) of main window texts.

        This changes depending on whether restricted or free drivers are
        used/available, thus the UI should update it whenever it changes a
        handler.
        '''
        proprietary_in_use = False
        proprietary_available = False

        for h_id in self.backend().available(self.argv_options.mode):
            info = self.backend().handler_info(h_id)
            #print 'main_window_text: info for', h_id, info
            if not bool(info['free']):
                proprietary_available = True
                if bool(info['used']):
                    proprietary_in_use = True
                    break

        if proprietary_in_use:
            heading = self._('Proprietary drivers are being used to make '
                    'this computer work properly.')
        else:
            heading = self._('No proprietary drivers are in use on this system.')

        if proprietary_available:
            subtext = self._(
            # %(os)s stands for the OS name. Prefix it or suffix it,
            # but do not replace it.
            'Proprietary drivers do not have public source code that %(os)s '
            'developers are free to modify. They represent a risk to you '
            'because they are only available on the types of computer chosen by '
            'the manufacturer, and security updates to them depend solely on the '
            'responsiveness of the manufacturer. %(os)s cannot fix or improve '
            'these drivers.') % {'os': OSLib.inst.os_vendor}
        else:
            subtext = ''

        return (heading, subtext)

    def get_handler_category(self, handler_id):
        '''Return string for handler category.'''

        if handler_id.startswith('xorg:'):
            return self._('Graphics driver')
        elif handler_id.startswith('firmware:'):
            return self._('Firmware')
        else:
            return self._('Device driver')

    def get_handler_tooltip(self, handler_id):
        '''Format handler rationale as a tooltip.

        Return None if the handler is None or does not have a rationale.
        '''
        try:
            info = polkit_auth_wrapper(self.backend().handler_info, handler_id)
        except UnknownHandlerException:
            return None
        try:
            r = info['rationale']
        except KeyError:
            # generate default rationale for some driver types
            if handler_id.startswith('firmware:'):
                fw_append = self._('Your hardware will not work without the firmware.')
                if bool(info['free']):
                    r = self._('The driver itself is already installed, but '
                        'it requires a piece of firmware which is not shipped '
                        'with the operating system.') + ' ' + fw_append
                else:
                    r = self._('While this driver itself is free software, it '
                        'relies on proprietary firmware which cannot be '
                        'legally shipped with the operating system.') + ' ' + fw_append
            else:
                if not bool(info['free']):
                    r = self._('This driver is necessary to support the '
                        'hardware, there is no free/open alternative.\n\n'
                        'If this driver is not enabled, the hardware will not '
                        'function.')
                else:
                    return None

        # opportunistic translation (shipped example drivers have translations)
        r = self._(r)

        tip = ''
        for par in r.split('\n'):
            if tip:
                tip += '\n'
            tip += '\n'.join(textwrap.wrap(par, 60))
        return tip

    def parse_argv(self):
        '''Parse command line arguments, and return (options, args) pair.'''

        # --check can have an optional numeric argument which sleeps for the
        # given number of seconds; this is mostly useful for the XDG autostart
        # .desktop file, to not do expensive operations right at session start
        def check_option_callback(option, opt_str, value, parser):
            if len(parser.rargs) > 0 and parser.rargs[0].isdigit():
                setattr(parser.values, 'check', int(parser.rargs.pop(0)))
            else:
                setattr(parser.values, 'check', 0)

        parser = optparse.OptionParser()
        parser.set_defaults(check=None)
        parser.add_option ('-c', '--check', action='callback',
                callback=check_option_callback,
                help=self._('Check for newly used or usable drivers and notify the user.'))
        parser.add_option ('-u', '--update-db', action='store_true',
                dest='update_db', default=False,
                help=self._('Query driver databases for newly available or updated drivers.'))
        parser.add_option ('-l', '--list', action='store_true',
                dest='list', default=False,
                help=self._('List available drivers and their status.'))
        parser.add_option ('--hardware-ids', action='store_true',
                dest='list_hwids', default=False,
                help=self._('List hardware identifiers from this system.'))
        parser.add_option ('-e', '--enable', type='string',
                dest='enable', default=None, metavar='DRIVER',
                help=self._('Enable a driver'))
        parser.add_option ('-d', '--disable', type='string',
                dest='disable', default=None, metavar='DRIVER',
                help=self._('Disable a driver'))
        parser.add_option ('--confirm', action='store_true',
                dest='confirm', default=False,
                help=self._('Ask for confirmation for --enable/--disable'))
        parser.add_option ('-C', '--check-composite', action='store_true',
                dest='check_composite', default=False,
                help=self._('Check if there is a graphics driver available that supports composite and offer to enable it'))
        parser.add_option ('-m', '--mode',
                type='choice', dest='mode', default='any',
                choices=['free', 'nonfree', 'any'],
                metavar='free|nonfree|any',
                help=self._('Only manage free/nonfree drivers. By default, all'
                ' available drivers with any license are presented.'))
        parser.add_option ('--dbus-server', action='store_true',
                dest='dbus_server', default=False,
                help=self._('Run as session D-BUS server.'))
        #parser.add_option ('--debug', action='store_true',
        #        dest='debug', default=False,
        #        help=self._('Enable debugging messages.'))

        (opts, args) = parser.parse_args()

        return (opts, args)

    def run(self):
        '''Evaluate command line arguments and do the appropriate action.

        If no argument was specified, this starts the interactive UI.
        
        This returns the exit code of the program.
        '''
        show_gui = True
        if self.argv_options.update_db:
            polkit_auth_wrapper(self.backend().update_driverdb)
            show_gui = False

        if self.argv_options.list:
            self.list()
            return 0
        elif self.argv_options.list_hwids:
            self.list_hwids()
            return 0
        elif self.argv_options.enable:
            if self.change_enable(self.argv_options.enable, True,
                self.argv_options.confirm):
                return 0
            else:
                return 1
        elif self.argv_options.disable:
            if self.change_enable(self.argv_options.disable, False,
                self.argv_options.confirm):
                return 0
            else:
                return 1
        elif self.argv_options.check is not None:
            if self.check():
                return 0
            else:
                return 1
        elif self.argv_options.check_composite:
            if self.check_composite():
                return 0
            else:
                return 1
        elif self.argv_options.dbus_server:
            self.dbus_server()
            return 0

        # start the UI
        if show_gui:
            self.ui_init()
            return self.ui_main_loop()
        else:
            return 0

    def list(self):
        '''Print a list of available handlers and their status to stdout.'''

        for h_id in self.backend().available(self.argv_options.mode):
            i = self.backend().handler_info(h_id)
            print '%s - %s (%s, %s, %s)' % (
                h_id, i['name'],
                bool(i['free']) and self.string_free or self.string_restricted,
                bool(i['enabled']) and self.string_enabled or self.string_disabled,
                bool(i['used']) and self.string_in_use or self.string_not_in_use)

    def list_hwids(self):
        '''Print a list of available handlers and their status to stdout.'''

        for h_id in self.backend().get_hardware():
            print h_id

    def change_enable(self, h_id, enable, confirm):
        '''Enable or disable a driver ID.'''

        try:
            i = polkit_auth_wrapper(self.backend().handler_info, h_id)
        except UnknownHandlerException:
            self.error_msg('%s: %s' % (self.string_unknown_driver, h_id))
            self.error_msg(self._('Use --list to see available drivers'))
            return False

        if 'can_change' in i:
            reason = i['can_change']
            self.error_msg(self._(reason))
            return False

        if bool(i['enabled']) == enable:
            return True

        try:
            if confirm:
                self.toggle_handler(h_id)
            else:
                polkit_auth_wrapper(self.backend().set_enabled, h_id, enable)
        except PermissionDeniedByPolicy:
            self.error_msg(self.string_unprivileged)
            return False

        return enable == bool(self.backend().handler_info(h_id)['enabled'])

    def check(self):
        '''Notify the user about newly used or available drivers since last check().
        
        Return True if any new driver is available which is not yet enabled.
        '''

        try:
            (new_used, new_avail) = polkit_auth_wrapper(
                self.backend().new_used_available, self.argv_options.mode)
        except PermissionDeniedByPolicy:
            self.error_msg(self.string_unprivileged)
            return False

        # any new restricted drivers?
        restricted_available = False
        for h_id in new_avail:
            if not bool(self.backend().handler_info(h_id)['free']):
                restricted_available = True
                break

        # throw out newly used free drivers; no need for education here
        for h_id in new_used + []: # create copy for iteration
            if bool(self.backend().handler_info(h_id)['free']):
                new_used.remove(h_id)

        notified = False

        # launch notifications if anything remains
        if new_avail:
            if restricted_available:
                self.ui_notification(self._('Restricted drivers available'),
                    self._('In order to use your hardware more efficiently, you'
                    ' can enable drivers which are not free software.'))
            else:
                self.ui_notification(self._('New drivers available'),
                    self._('There are new or updated drivers available for '
                    'your hardware.'))
            notified = True
        elif new_used:
            self.ui_notification(self._('New restricted drivers in use'),
                # %(os)s stands for the OS name. Prefix it or suffix it,
                # but do not replace it.
                self._('In order for this computer to function properly, %(os)s is '
                'using driver software that cannot be supported by %(os)s.') % 
                    {'os': OSLib.inst.os_vendor})
            notified = True

        if notified:
            # we need to stay in the main loop so that the tray icon stays
            # around
            self.ui_main_loop()

        return len(new_avail) > 0

    def check_composite(self):
        '''Check for a composite-enabling X.org driver.

        If one is available and not installed, offer to install it and return
        True if installation succeeded. Otherwise return False.
        '''

        h_id = self.backend().check_composite()

        if h_id:
            self.toggle_handler(h_id)
            return bool(self.backend().handler_info(h_id)['enabled'])

        self.error_msg(self._('There is no available graphics driver for your system which supports the composite extension, or the current one already supports it.'))
        return False

    def _install_progress_handler(self, phase, cur, total):
        if not self._install_progress_shown:
            self.ui_progress_start('', 
                self._('Downloading and installing driver...'), total)
            self._install_progress_shown = True
        self.ui_progress_update(cur, total)
        self.ui_idle()

    def _remove_progress_handler(self, cur, total):
        if not self._install_progress_shown:
            self.ui_progress_start('', 
                self._('Removing driver...'), total)
            self._install_progress_shown = True
        self.ui_progress_update(cur, total)
        self.ui_idle()

    def toggle_handler(self, handler_id):
        '''Callback for toggling the handler enable/disable state in the UI.
        
        After this, you need to refresh the UI's handler tree view if this
        method returns True.
        '''

        i = self.backend().handler_info(handler_id)

        # check if we can change at all
        if 'can_change' in i:
            self.error_message(self._('Cannot change driver'), i['can_change'])
            return False

        en = bool(i['enabled'])

        # construct and ask confirmation question
        if en:
            title = self._('Disable driver?')
            action = self._('_Disable', True)
        else:
            title = self._('Enable driver?')
            action = self._('_Enable', True)

        d = i.get('description', '')
        r = i.get('rationale', '')
        # opportunistic translation (shipped example drivers have translations)
        if d:
            d = self._(d)
        if r:
            r = self._(r)

        if d and r:
            subtext = d.strip() + '\n\n' + r
        elif d:
            subtext = d
        elif r:
            subtext = r
        else:
            subtext = None
        n = i['name']
        if not self.confirm_action(title, self._(n), subtext, action):
            return False

        # go
        # TODO: when being called with --confirm --enable, this needs partial
        # ui_init() for loading the UI
        try:
            self._install_progress_shown = False
            polkit_auth_wrapper(dbus_sync_call_signal_wrapper, 
                self.backend(), 'set_enabled',
                {'install_progress': self._install_progress_handler, 
                 'remove_progress': self._remove_progress_handler}, 
                handler_id, not en)
            if self._install_progress_shown:
                self.ui_progress_finish()
        except PermissionDeniedByPolicy:
            self.error_message('', self.string_unprivileged)
            return False

        return True

    def download_url(self, url, filename=None, data=None):
        '''Download an URL into a local file, and display a progress dialog.
        
        If filename is not given, a temporary file will be created.

        Additional POST data can be submitted for HTTP requests in the data
        argument (see urllib2.urlopen).

        Return (filename, headers) tuple, or (None, headers) if the user
        cancelled the download.
        '''
        block_size = 8192
        current_size = 0
        try:
            f = urllib2.urlopen(url)
        except Exception, e:
            self.error_message(self._('Download error'), str(e))
            return (None, None)
        headers = f.info()

        if 'Content-Length' in headers:
            total_size = int(headers['Content-Length'])
        else:
            total_size = -1

        self.ui_progress_start(self.string_download_progress_title, url,
            total_size)

        if filename:
            tfp = open(filename, 'wb')
            result_filename = filename
        else:
            (fd, result_filename) = tempfile.mkstemp()
            tfp = os.fdopen(fd, 'wb')

        try:
            while current_size < total_size:
                block = f.read(block_size)
                tfp.write (block)
                current_size += len(block)
                # if True, user canceled download
                if self.ui_progress_update(current_size, total_size):
                    # if we created a temporary file, clean it up
                    if not filename:
                        os.unlink(result_filename)
                    result_filename = None
                    break
        finally:
            tfp.close()
            f.close()
            self.ui_progress_finish()

        return (result_filename, headers)

    @classmethod
    def error_msg(klass, msg):
        '''Print msg to stderr, and intercept IOErrors.'''

        try:
            print >> sys.stderr, msg
        except IOError:
            pass

    #
    # Session D-BUS server methods
    # 

    DBUS_INTERFACE_NAME = 'com.ubuntu.DeviceDriver'

    def dbus_server(self):
        '''Run session D-BUS server backend.'''

        self.backend().add_driverdb('OpenPrintingDriverDB', []) #XXX

        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
        bus = dbus.SessionBus()
        dbus_name = dbus.service.BusName(DBUS_BUS_NAME, bus)

        dbus.service.Object.__init__(self, bus, '/GUI')
        self.dbus_server_main_loop = gobject.MainLoop()
        self.dbus_server_main_loop.run()

    @classmethod
    def get_dbus_client(klass):
        '''Return a dbus.Interface object for the server.'''

        obj = dbus.SessionBus().get_object(DBUS_BUS_NAME, '/GUI')
        return dbus.Interface(obj, AbstractUI.DBUS_INTERFACE_NAME)

    @dbus.service.method(DBUS_INTERFACE_NAME,
        in_signature='s', out_signature='b', sender_keyword='sender',
        connection_keyword='conn')
    def search_driver(self, hwid, sender=None, conn=None):
        '''Search configured driver DBs for a particular hardware component.

        The hardware component is described as HardwareID type and value,
        separated by a colon. E. g.  "modalias:pci:12345" or
        "printer_deviceid:MFG:FooTech;MDL:X-12;CMD:GDI". This searches the
        enabled driver databases for a matching driver. If it finds one, it
        offers it to the user. This returns True if a driver was found,
        acknowledged by the user, and installed, otherwise it returns False.
        '''
        drivers = self.backend().search_driver(hwid)
        result = False

        if drivers:
            result = self.change_enable(drivers[0], True, True)

        # we are D-BUS activated, so let's free resources early
        if self.dbus_server_main_loop:
            self.dbus_server_main_loop.quit()

        return result

    #
    # The following methods must be implemented in subclasses
    # 

    def convert_keybindings(self, str):
        '''Convert keyboard accelerators to the particular UI's format.

        The abstract UI and drivers use the '_' prefix to mark a keyboard
        accelerator.

        A double underscore ('__') is converted to a real '_'.'''

        raise NotImplementedError, 'subclasses need to implement this'

    def ui_init(self):
        '''Initialize UI.
        
        This should set up presentation of handlers and show the main
        window.
        '''
        raise NotImplementedError, 'subclasses need to implement this'

    def ui_main_loop(self):
        '''Main loop for the user interface.
        
        This should return if the user wants to quit the program, and return
        the exit code.
        '''
        raise NotImplementedError, 'subclasses need to implement this'

    def error_message(self, title, text):
        '''Present an error message box.'''

        raise NotImplementedError, 'subclasses need to implement this'

    def confirm_action(self, title, text, subtext=None, action=None):
        '''Present a confirmation dialog.

        If action is given, it is used as button label instead of the default
        'OK'.  Return True if the user confirms, False otherwise.
        '''
        raise NotImplementedError, 'subclasses need to implement this'

    def ui_notification(self, title, text):
        '''Present a notification popup.

        This should preferably create a tray icon. Clicking on the tray icon or
        notification should run the GUI.
        '''
        raise NotImplementedError, 'subclasses need to implement this'

    def ui_idle(self):
        '''Process pending UI events and return.

        This is called while waiting for external processes such as package
        installers.
        '''
        raise NotImplementedError, 'subclasses need to implement this'

    def ui_progress_start(self, title, description, total):
        '''Create a progress dialog.'''

        raise NotImplementedError, 'subclasses need to implement this'

    def ui_progress_update(self, current, total):
        '''Update status of current progress dialog.
        
        current/total specify the number of steps done and total steps to
        do, or -1 if it cannot be determined. In this case the dialog should
        display an indeterminated progress bar (bouncing back and forth).

        This should return True to cancel, and False otherwise.
        '''
        raise NotImplementedError, 'subclasses need to implement this'

    def ui_progress_finish(self):
        '''Close the current progress dialog.'''

        raise NotImplementedError, 'subclasses need to implement this'
